from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from typing import Generator, List, Union

import pandas as pd

from kabutobashi.domain.values import IStockRecordsetRepository, StockRecordset
from kabutobashi.infrastructure.crawler import crawl_multiple

__all__ = ["IStockRecordsetStorageRepository", "StockRecordsetStorageBasicRepository", "StockRecordsetCrawler"]


class IStockRecordsetStorageRepository(IStockRecordsetRepository):
    def __init__(self, use_mp: bool, max_workers: int):
        self.use_mp = use_mp
        self.max_workers = max_workers

    @abstractmethod
    def _read_path_generator(self) -> Generator[str, None, None]:
        raise NotImplementedError()

    def _stock_recordset_read(self) -> StockRecordset:
        df_list = []
        if self.use_mp:
            with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                map_gen = executor.map(pd.read_csv, self._read_path_generator())
                for response in map_gen:
                    df_list.append(response)
        else:
            df_list = [pd.read_csv(p) for p in self._read_path_generator()]

        df = pd.concat(df_list)
        return StockRecordset.of(df=df)

    @abstractmethod
    def _write_path_generator(self) -> Generator[str, None, None]:
        raise NotImplementedError()

    def _stock_recordset_write(self, data: StockRecordset):
        df = data.to_df(minimum=False)
        for p in self._write_path_generator():
            df.to_csv(p, index=False)


class StockRecordsetStorageBasicRepository(IStockRecordsetStorageRepository):
    def __init__(self, path_candidate: Union[str, list], use_mp: bool = False, max_workers: int = 2):
        super().__init__(use_mp=use_mp, max_workers=max_workers)
        self.path_candidate = path_candidate

    def _read_path_generator(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()

    def _write_path_generator(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()


class StockRecordsetCrawler:
    def __init__(self, use_mp: bool = False, max_workers: int = 2):
        self.use_mp = use_mp
        self.max_workers = max_workers

    def get(self, code_list: list, dt: str) -> StockRecordset:
        # 日次の株データ取得
        stock_data: List[dict] = crawl_multiple(code_list=code_list, max_workers=self.max_workers, dt=dt)

        # データを整形してStockDataとして保存
        df = pd.DataFrame(stock_data)
        return StockRecordset.of(df=df)



from dataclasses import dataclass, field
from typing import Generator, Tuple

import pandas as pd

from kabutobashi.domain.errors import KabutobashiEntityError

from .stock_recordset import StockRecordset

__all__ = ["StockDataSingleCode"]


@dataclass(frozen=True)
class StockDataSingleCode:
    """
    StockDataSingleCode: ValueObject

    Examples:
        >>> import kabutobashi as kb
        >>> import pandas as pd
        >>> data_list = []
        >>> records = kb.example()
        >>> parameterize_methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        >>>
        >>> for df in records.to_code_iterable(until=1, row_more_than=80):
        >>>     sdsc = StockDataSingleCode.of(df=df)
        >>>     for idx, df_x, df_y in sdsc.sliding_split():
        >>>         df_params = kb.StockDataProcessedBySingleMethod.of(df=df_x, methods=parameterize_methods)
        >>>         # diff:= df_y.last - df_x.last
        >>>         start = list(df_x["close"])[-1]
        >>>         end = list(df_y["close"])[-1]
        >>>         diff = end - start
        >>>         d = df_params.get_parameters()
        >>>         d.update({"diff": diff})
        >>>         data_list.append(d)
        >>>  data_for_ml = pd.DataFrame(data_list)
    """

    code: str = field(metadata={"jp": "銘柄コード"})
    stop_updating: bool = field(metadata={"jp": "更新停止"})
    contains_outlier: bool = field(metadata={"jp": "例外値を含む"})
    stock_recordset: StockRecordset = field(metadata={"jp": "株データ"})
    len_: int = field(metadata={"jp": "データ数"})

    def __post_init__(self):
        self._code_constraint_check(stock_recordset=self.stock_recordset)

    @staticmethod
    def _code_constraint_check(stock_recordset: StockRecordset):
        brands = stock_recordset.get_code_list()
        if len(brands) > 1:
            raise KabutobashiEntityError("multiple code")
        elif len(brands) == 0:
            raise KabutobashiEntityError("no code")
        code_records = list(set([v.code for v in stock_recordset.recordset]))
        if len(code_records) > 1:
            raise KabutobashiEntityError("multiple code")
        elif len(code_records) == 0:
            raise KabutobashiEntityError("no code")

    @staticmethod
    def of(df: pd.DataFrame):
        recordset = StockRecordset.of(df=df)

        # codeの確認
        StockDataSingleCode._code_constraint_check(stock_recordset=recordset)
        code = recordset.get_code_list()[0]
        return StockDataSingleCode(
            code=code,
            stock_recordset=recordset,
            stop_updating=StockDataSingleCode._check_recent_update(df=df),
            contains_outlier=any([v.is_outlier() for v in recordset.recordset]),
            len_=len(recordset.recordset),
        )

    @staticmethod
    def _check_recent_update(df: pd.DataFrame) -> bool:
        """
        直近の更新が止まっているかどうか
        """
        return (
            (len(df["open"].tail(10).unique()) == 1)
            or (len(df["high"].tail(10).unique()) == 1)
            or (len(df["low"].tail(10).unique()) == 1)
            or (len(df["close"].tail(10).unique()) == 1)
        )

    def sliding_split(
        self, *, buy_sell_term_days: int = 5, sliding_window: int = 60, step: int = 3
    ) -> Generator[Tuple[int, pd.DataFrame, pd.DataFrame], None, None]:
        """
        単一の銘柄に関してwindow幅を ``sliding_window`` 日として、
        保持しているデータの期間の間をslidingしていく関数。

        Args:
            buy_sell_term_days: この日数後までデータを切り出す。
            sliding_window: slidingしていくwindow幅
            step: windowsをずらしていく期間

        Returns:
            idx: 切り出された番号。
            df_for_x: 特徴量を計算するためのDataFrame。
            df_for_y: ``buy_sell_term_days`` 後のDataFrameを返す。値動きを追うため。
        """
        df = self.to_df()
        df_length = len(df.index)
        if df_length < buy_sell_term_days + sliding_window:
            raise KabutobashiEntityError("入力されたDataFrameの長さがwindow幅よりも小さいです")
        loop = df_length - (buy_sell_term_days + sliding_window)
        for idx, i in enumerate(range(0, loop, step)):
            offset = i + sliding_window
            end = offset + buy_sell_term_days
            yield idx, df[i:offset], df[offset:end]

    def to_df(self, minimum=True, latest=False) -> pd.DataFrame:
        return self.stock_recordset.to_df(minimum=minimum, latest=latest)

    def __len__(self):
        return self.len_



from .html_pages import StockInfoHtmlPage, StockIpoHtmlPage, StockWeeks52HighLowHtmlPage
from .stock_data_estimated import StockDataEstimatedBySingleFilter
from .stock_data_processed import StockDataProcessedBySingleMethod
from .stock_data_single_code import StockDataSingleCode
from .stock_data_visualized import StockDataVisualized
from .stock_recordset import IStockRecordsetRepository, StockRecordset



import functools
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from typing import List, Union

from kabutobashi.domain.errors import KabutobashiPageError
from kabutobashi.domain.services import StockInfoHtmlDecoder
from kabutobashi.domain.values import StockInfoHtmlPage

logger = getLogger(__name__)


def crawl_single(code: Union[int, str], dt: str) -> dict:
    try:
        stock_page_html = StockInfoHtmlPage.of(code=code, dt=dt)
        result = StockInfoHtmlDecoder(html_page=stock_page_html).decode()
        return result
    except KabutobashiPageError:
        return {}
    except AttributeError:
        logger.exception(f"error occurred at: {code}")
        return {}
    except Exception:
        logger.exception(f"error occurred at: {code}")
        return {}


def crawl_multiple(code_list: List[Union[int, str]], dt: str, max_workers: int = 2) -> List[dict]:
    partial_crawl_single = functools.partial(crawl_single, dt=dt)
    response_list = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        map_gen = executor.map(partial_crawl_single, code_list)
        for response in map_gen:
            response_list.append(response)
    return response_list



import pandas as pd
import pydantic
import pytest

import kabutobashi as kb
from kabutobashi.domain.errors import KabutobashiEntityError


class TestStockBrand:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.StockBrand(
                id=None,
                code="1234",
                market="",
                name="",
                unit="",
                market_capitalization="",
                industry_type="",
                issued_shares="",
            )


class TestStockRecord:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.StockRecord(
                id=None,
                code="1234",
                open="",
                high="",
                low="",
                close="",
                psr="",
                per="",
                pbr="",
                volume="",
                dt="",
            )


class TestStockIpo:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.StockIpo(
                id=None, code="", manager="", stock_listing_at="", public_offering="", evaluation="", initial_price=""
            )


class TestWeeks52HihLow:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.Weeks52HighLow(
                code="", brand_name="", close="", buy_or_sell="", volatility_ratio="", volatility_value=""
            )


class TestStockDataSingleCode:
    def test_of(self, data_path):
        df = pd.read_csv(f"{data_path}/example.csv.gz")
        df["code"] = df["code"].astype(str)
        single_code = df[df["code"] == "1375"]
        _ = kb.StockDataSingleCode.of(df=single_code)

        # check None
        with pytest.raises(KabutobashiEntityError):
            _ = kb.StockDataSingleCode(
                code="-",
                stock_recordset=kb.StockRecordset.of(df=pd.DataFrame()),
                stop_updating=False,
                contains_outlier=False,
                len_=0,
            )

        # check multiple code
        with pytest.raises(KabutobashiEntityError):
            _ = kb.StockDataSingleCode.of(df=df)

        # check invalid column
        with pytest.raises(KabutobashiEntityError):
            _ = kb.StockDataSingleCode.of(df=single_code[["close"]])

    def test_get_df(self, data_path):
        df = pd.read_csv(f"{data_path}/example.csv.gz")
        df["code"] = df["code"].astype(str)
        single_code = df[df["code"] == "1375"]
        sdsc = kb.StockDataSingleCode.of(df=single_code)

        required_cols = ["code", "open", "close", "high", "low", "volume", "per", "psr", "pbr", "dt"]
        optional_cols = ["name", "industry_type", "market", "unit"]

        # check minimum df
        minimum_df = sdsc.to_df()
        assert all([(c in minimum_df.columns) for c in required_cols])
        assert all([(c not in minimum_df.columns) for c in optional_cols])

        # check full df
        full_df = sdsc.to_df(minimum=False)
        assert all([(c in full_df.columns) for c in required_cols])
        assert all([(c in full_df.columns) for c in optional_cols])

        latest_date_df = sdsc.to_df(latest=True)
        assert len(latest_date_df.index) == 1


class TestStockRecordset:
    def test_code_iterable(self):
        records = kb.example()
        for _ in records.to_code_iterable(until=1):
            pass


class TestStockSingleAggregate:
    def test_pass(self):
        records = kb.example()
        methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        agg = kb.StockCodeSingleAggregate.of(entity=records, code="1375")
        _ = agg.with_processed(methods=methods).with_estimated(estimate_filters=kb.estimate_filters)



from dataclasses import dataclass, field
from typing import Any, Dict, List

import pandas as pd
from cerberus import Validator

from kabutobashi.domain.errors import KabutobashiEntityError


@dataclass(frozen=True)
class StockDataProcessedBySingleMethod:
    """
    StockDataProcessedBySingleMethod: ValueObject
    Holds data processed by singular-Method.
    """

    code: str
    start_at: str
    end_at: str
    applied_method_name: str
    df: pd.DataFrame = field(repr=False)
    df_required_columns: List[str] = field(repr=False)
    parameters: Dict[str, Any]
    color_mapping: list = field(repr=False)
    visualize_option: dict = field(repr=False)
    COLOR_MAPPING_SCHEMA = {
        "df_key": {"type": "string"},
        "color": {"type": "string"},
        "label": {"type": "string"},
        "plot": {"type": "string", "allowed": ["plot", "bar"], "required": False},
    }
    VISUALIZE_OPTION_SCHEMA = {"position": {"type": "string", "allowed": ["in", "lower", "-"]}}

    def __post_init__(self):
        self._check_color_mapping(data=self.color_mapping)
        self._check_visualize_option(data=self.visualize_option)

    def _check_color_mapping(self, data: list):
        validator = Validator(self.COLOR_MAPPING_SCHEMA)
        for d in data:
            if not validator.validate(d):
                raise KabutobashiEntityError(validator)

    def _check_visualize_option(self, data: dict):
        validator = Validator(self.VISUALIZE_OPTION_SCHEMA)
        if not validator.validate(data):
            raise KabutobashiEntityError(validator)

    def get_impact(self, influence: int = 2, tail: int = 5) -> Dict[str, float]:
        """

        Args:
            influence:
            tail:

        Returns:
            Dict[str, float]

        Examples:
        """
        return {self.applied_method_name: self._get_impact(df=self.df, influence=influence, tail=tail)}

    @staticmethod
    def _get_impact(df: pd.DataFrame, influence: int, tail: int) -> float:
        """
        売りと買いのシグナルの余波の合計値を返す。

        Args:
            df:
            influence:
            tail:

        Returns:
            [-1,1]の値をとる。-1: 売り、1: 買いを表す
        """
        df["buy_impact"] = df["buy_signal"].ewm(span=influence).mean()
        df["sell_impact"] = df["sell_signal"].ewm(span=influence).mean()
        buy_impact_index = df["buy_impact"].iloc[-tail:].sum()
        sell_impact_index = df["sell_impact"].iloc[-tail:].sum()
        return round(buy_impact_index - sell_impact_index, 5)



from dataclasses import dataclass


@dataclass(frozen=True)
class StockDataEstimatedBySingleFilter:
    """
    StockDataEstimatedBySingleFilter: ValueObject
    """

    code: str
    estimated_value: float
    estimate_filter_name: str

    def weighted_estimated_value(self, weights: dict) -> float:
        weight = weights.get(self.estimate_filter_name, 1)
        return weight * self.estimated_value



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, NoReturn, Optional, Union

import pandas as pd

from kabutobashi.domain.errors import KabutobashiEntityError
from kabutobashi.domain.services.estimate_filter import EstimateFilter
from kabutobashi.domain.services.method import Method
from kabutobashi.domain.values import (
    StockDataEstimatedBySingleFilter,
    StockDataProcessedBySingleMethod,
    StockDataSingleCode,
    StockDataVisualized,
    StockRecordset,
)

__all__ = ["StockCodeSingleAggregate", "IStockCodeSingleAggregateRepository"]


@dataclass(frozen=True)
class StockCodeSingleAggregate:
    """
    StockCodeSingleAggregate: Aggregate

    Examples:
        >>> import kabutobashi as kb
        >>> import pandas as pd
        >>> data_list = []
        >>> records = kb.example()
        >>> methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        >>> filters = kb.estimate_filters
        >>>
        >>> for df in records.to_code_iterable(until=1, row_more_than=80):
        >>>     agg = kb.StockCodeSingleAggregate.of(entity=df).with_processed(methods).with_estimated(filters)
        >>>     print(agg.weighted_estimated_value())
    """

    code: str
    single_code: StockDataSingleCode
    processed_list: List[StockDataProcessedBySingleMethod] = field(default_factory=list, repr=False)
    estimated_list: List[StockDataEstimatedBySingleFilter] = field(default_factory=list, repr=False)

    @staticmethod
    def of(
        entity: Union[pd.DataFrame, StockDataSingleCode, StockRecordset], *, code: Optional[str] = None
    ) -> "StockCodeSingleAggregate":
        if type(entity) is pd.DataFrame:
            single_code = StockDataSingleCode.of(df=entity)
        elif type(entity) is StockDataSingleCode:
            single_code = entity
        elif type(entity) is StockRecordset:
            if code is None:
                raise KabutobashiEntityError("code is required")
            df_ = entity.to_df(code=code)
            single_code = StockDataSingleCode.of(df=df_)
        else:
            raise KabutobashiEntityError("accept pd.DataFrame or StockDataSingleCode")
        return StockCodeSingleAggregate(code=single_code.code, single_code=single_code)

    def _to_single_processed(self, method: Method) -> StockDataProcessedBySingleMethod:
        df = self.single_code.to_df()
        # 日時
        start_at = list(df["dt"])[0]
        end_at = list(df["dt"])[-1]

        # 必要なパラメータの作成
        columns = ["dt", "open", "close", "high", "low", "buy_signal", "sell_signal"] + method.processed_columns()
        df_p = df.pipe(method.method).pipe(method.signal).loc[:, columns]
        params = method.parameterize(df_x=df, df_p=df_p)

        return StockDataProcessedBySingleMethod(
            code=self.code,
            start_at=start_at,
            end_at=end_at,
            applied_method_name=method.method_name,
            df=df_p,
            df_required_columns=columns,
            parameters=params,
            color_mapping=method.color_mapping(),
            visualize_option=method.visualize_option(),
        )

    def with_processed(self, methods: List[Method]) -> "StockCodeSingleAggregate":
        # check all methods
        for method in methods:
            if not isinstance(method, Method):
                raise KabutobashiEntityError()

        return StockCodeSingleAggregate(
            code=self.code, single_code=self.single_code, processed_list=[self._to_single_processed(m) for m in methods]
        )

    def _to_single_estimated(self, estimate_filter: EstimateFilter) -> StockDataEstimatedBySingleFilter:
        def get_impacts(influence: int = 2, tail: int = 5) -> Dict[str, float]:
            data_ = {}
            for a in self.processed_list:
                data_.update(a.get_impact(influence=influence, tail=tail))
            return data_

        def get_parameters():
            data_ = {}
            for a in self.processed_list:
                data_.update(a.parameters)
            return data_

        data = {}
        data.update(get_impacts())
        data.update(get_parameters())
        return StockDataEstimatedBySingleFilter(
            code=self.code,
            estimate_filter_name=estimate_filter.estimate_filter_name,
            estimated_value=estimate_filter.estimate(data=data),
        )

    def with_estimated(self, estimate_filters: List[EstimateFilter]) -> "StockCodeSingleAggregate":
        """

        Returns:
            StockDataEstimatedByMultipleFilter
        """
        return StockCodeSingleAggregate(
            code=self.code,
            single_code=self.single_code,
            processed_list=self.processed_list,
            estimated_list=[self._to_single_estimated(estimate_filter=ef) for ef in estimate_filters],
        )

    def weighted_estimated_value(self, weights: dict = None) -> float:
        if not weights:
            weights = {}
        return sum([e.weighted_estimated_value(weights=weights) for e in self.estimated_list])

    def estimate_filter_concat_name(self) -> str:
        estimate_filter_names = sorted([e.estimate_filter_name for e in self.estimated_list])
        return "_".join(estimate_filter_names)

    def visualize(self, size_ratio: int = 2) -> StockDataVisualized:
        if not self.processed_list:
            raise KabutobashiEntityError("call with_processed() before.")
        return StockDataVisualized.of(processed=self.processed_list, size_ratio=size_ratio)


class IStockCodeSingleAggregateRepository(metaclass=ABCMeta):
    def read(self, code: str) -> "StockCodeSingleAggregate":
        return StockCodeSingleAggregate(
            code=code,
            single_code=self._stock_data_read(code=code),
            processed_list=self._stock_processed_read(code=code),
            estimated_list=self._stock_estimated_read(code=code),
        )

    @abstractmethod
    def _stock_data_read(self, code: str) -> "StockDataSingleCode":
        raise NotImplementedError()

    @abstractmethod
    def _stock_processed_read(self, code: str) -> List["StockDataProcessedBySingleMethod"]:
        raise NotImplementedError()

    @abstractmethod
    def _stock_estimated_read(self, code: str) -> List["StockDataEstimatedBySingleFilter"]:
        raise NotImplementedError()

    def write(self, data: StockCodeSingleAggregate) -> NoReturn:
        self._stock_data_write(data=data.single_code)
        self._stock_processed_write(data=data.processed_list)
        self._stock_estimated_write(data=data.estimated_list)

    @abstractmethod
    def _stock_data_write(self, data: StockDataSingleCode) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def _stock_processed_write(self, data: List[StockDataProcessedBySingleMethod]) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def _stock_estimated_write(self, data: List[StockDataEstimatedBySingleFilter]) -> NoReturn:
        raise NotImplementedError()



from dataclasses import dataclass, field
from typing import List

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from mplfinance.original_flavor import candlestick_ohlc

from kabutobashi.domain.errors import KabutobashiEntityError
from kabutobashi.domain.values.stock_data_processed import StockDataProcessedBySingleMethod


@dataclass(frozen=True)
class StockDataVisualized:
    """
    StockDataVisualized: ValueObject
    Used to visualize.
    """

    fig: plt.Figure
    size_ratio: int
    processed: List[StockDataProcessedBySingleMethod] = field(default_factory=list)

    @staticmethod
    def of(processed: List[StockDataProcessedBySingleMethod], size_ratio: int = 2):
        return StockDataVisualized(
            fig=StockDataVisualized._visualize(processed_list=processed, size_ratio=size_ratio), size_ratio=size_ratio
        )

    @staticmethod
    def _add_ax_candlestick(ax, _df: pd.DataFrame):
        # datetime -> float
        time_series = mdates.date2num(_df["dt"])
        data = _df[["open", "high", "low", "close"]].values.T
        # data
        ohlc = np.vstack((time_series, data)).T
        candlestick_ohlc(ax, ohlc, width=0.7, colorup="g", colordown="r")

    @staticmethod
    def _visualize(processed_list: List[StockDataProcessedBySingleMethod], size_ratio: int):
        """
        Visualize Stock Data.

        Args:
            processed_list:
            size_ratio: determine the size of the graph, default 2.

        Returns:
            Figure
        """

        def _n_rows() -> int:
            lower_nums = len([p for p in processed_list if p.visualize_option["position"] == "lower"])
            return 1 + lower_nums

        n_rows = _n_rows()

        def _gridspec_kw() -> dict:
            if n_rows == 1:
                return {"height_ratios": [3]}
            return {"height_ratios": [3] + [1] * (n_rows - 1)}

        gridspec_kw = _gridspec_kw()
        fig, axs = plt.subplots(
            nrows=n_rows, ncols=1, figsize=(6 * size_ratio, 5 * size_ratio), gridspec_kw=gridspec_kw
        )
        # auto-formatting x-axis
        fig.autofmt_xdate()

        # set candlestick base
        base_df = processed_list[0].df[["dt", "open", "close", "high", "low"]]
        StockDataVisualized._add_ax_candlestick(axs[0], base_df)

        ax_idx = 1
        # plots
        for processed in processed_list:
            position = processed.visualize_option["position"]
            df = processed.df
            time_series = mdates.date2num(df["dt"])
            mapping = processed.color_mapping
            if position == "in":
                for m in mapping:
                    df_key = m["df_key"]
                    color = m["color"]
                    label = m["label"]
                    axs[0].plot(time_series, df[df_key], label=label)
                # display labels
                axs[0].legend(loc="best")
            elif position == "lower":
                for m in mapping:
                    df_key = m["df_key"]
                    color = m["color"]
                    label = m["label"]
                    plot = m.get("plot", "plot")
                    if plot == "plot":
                        # type FloatingArray is no accepted ...
                        # so `df[df_key].astype(float)`
                        axs[ax_idx].plot(time_series, df[df_key].astype(float), label=label)
                    elif plot == "bar":
                        axs[ax_idx].bar(time_series, df[df_key], label=label)
                # display labels
                axs[ax_idx].legend(loc="best")
                # lower
                ax_idx += 1
            elif position == "-":
                # technical_analysis以外のmethodが入っている場合
                pass
            else:
                raise KabutobashiEntityError()

        return fig



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Generator, List, NoReturn, Optional, Set

import pandas as pd

from kabutobashi.domain.entity import OPTIONAL_COL, REQUIRED_COL, StockBrand, StockRecord
from kabutobashi.domain.errors import KabutobashiEntityError


@dataclass(frozen=True)
class StockRecordset:
    """
    StockRecordset: root-entity
    """

    brand_set: Set[StockBrand] = field(repr=False)
    recordset: List[StockRecord] = field(repr=False)

    def __post_init__(self):
        if not self.recordset:
            raise KabutobashiEntityError(f"required stock_data")

    @staticmethod
    def of(df: pd.DataFrame) -> "StockRecordset":
        if "code" not in df.columns:
            raise KabutobashiEntityError
        recordset = []
        brand_set = set()
        df = df.dropna(subset=["code"])
        for _, row in df.iterrows():
            recordset.append(StockRecord.loads(dict(row)))
            brand_set.add(StockBrand.loads(data=dict(row)))
        return StockRecordset(brand_set=brand_set, recordset=recordset)

    def get_code_list(self) -> List[str]:
        return list([v.code for v in self.brand_set])

    def _to_df(self, code: Optional[str]) -> pd.DataFrame:
        df_brand = pd.DataFrame([v.dumps() for v in self.brand_set])
        if code:
            df_brand = df_brand[df_brand["code"] == code]
        df_record = pd.DataFrame([v.dumps() for v in self.recordset])
        df = pd.merge(left=df_brand, right=df_record, how="inner", on="code")

        df = df.convert_dtypes()
        # order by dt
        idx = pd.to_datetime(df["dt"]).sort_index()
        df.index = idx
        df = df.sort_index()
        return df

    def to_df(self, *, minimum=True, latest=False, code: Optional[str] = None):
        df = self._to_df(code=code)

        if latest:
            latest_dt = max(df["dt"])
            df = df[df["dt"] == latest_dt]

        if minimum:
            return df[REQUIRED_COL]
        else:
            return df[REQUIRED_COL + OPTIONAL_COL]

    def to_single_code(self, code: str) -> "StockRecordset":
        if type(code) is not str:
            raise KabutobashiEntityError(f"code must be type of `str`")
        return StockRecordset.of(df=self._to_df(code=code))

    def to_code_iterable(
        self,
        until: Optional[int] = None,
        *,
        skip_reit: bool = True,
        row_more_than: Optional[int] = 80,
        code_list: list = None,
    ) -> Generator[pd.DataFrame, None, None]:
        _count = 0
        df = self._to_df(code=None)

        if code_list:
            df = df[df["code"].isin(code_list)]
        if skip_reit:
            df = df[~(df["market"] == "東証REIT")]

        for code, df_ in df.groupby("code"):
            if row_more_than:
                if len(df_.index) < row_more_than:
                    continue

            # add counter if yield
            if until:
                if _count >= until:
                    return
            _count += 1

            yield df_


class IStockRecordsetRepository(metaclass=ABCMeta):
    def read(self) -> "StockRecordset":
        return self._stock_recordset_read()

    @abstractmethod
    def _stock_recordset_read(self) -> "StockRecordset":
        raise NotImplementedError()

    def write(self, data: StockRecordset) -> NoReturn:
        self._stock_recordset_write(data=data)

    @abstractmethod
    def _stock_recordset_write(self, data: StockRecordset) -> NoReturn:
        raise NotImplementedError()
